'''
/***************************************************************************
Name       :  geowrsi_controller.py
Description:  Controller for GeoWRSI Tool
copyright  :  (C) 2022-2023 by FEWS
email      :  dhackman@contractor.usgs.gov
Author     :  Derek Hackman
Modified   :  mm/dd/yyyy - descripton - Description

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
'''
import os

from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import QDialog, QMessageBox

from qgis.core import Qgis, QgsMessageLog, QgsProject, QgsLayerTree
from qgis.utils import iface

from fews_tools.forms.Ui_GeoWRSIRun import Ui_RunGeoWRSI
from fews_tools.models.crops_model import CropsModel
from fews_tools.models.datasets_model import DatasetsModel
from fews_tools.models.region_model import RegionModel
from fews_tools.models.workspace_setup_model import WorkspaceSetupModel
from fews_tools.workers.geowrsi_worker import GeoWRSIWorker
from fews_tools.utilities.animation_utilities import setup_animation, set_start_end

from fews_tools import fews_tools_config as config
from fews_tools.utilities import geoclim_qgs_utilities as qgs_util
from fews_tools.utilities import geoclim_utilities as util
from fews_tools.utilities import logging_utilities as log_util
from fews_tools.utilities import geowrsi_utilities as geo_util


class GeoWRSIController (QDialog):
    '''
    Class for calculating the WRSI
    '''

    def __init__(self):
        QDialog.__init__(self)
        self.ui = Ui_RunGeoWRSI()
        self.ui.setupUi(self)
        self.is_processing = False  # Flag to indicate if running a process
        self.wrksp_setup = WorkspaceSetupModel()
        self.ds_dic_ppt = {}
        self.ds_dic_pet = {}
        self.region_info = RegionModel()
        self.reg_dic = {}
        self.crop_dic = {}
        self.input_file_path =\
            os.path.join(self.wrksp_setup.get_workspace(),
                         config.PROGRAM_SETTINGS,
                         config.GEOWRSI_SETTINGS_FILE)
        self.output_file_path =\
            self.wrksp_setup.get_output_path()
        self.set_dic = {
            # Region/crop tab
            'analysis_region': '',
            'crop_type': '',
            # SOS tab
            'sos_type': '',
            'sos_offset': 0,
            'sos_period': 1,
            'sos_file': '',
            # LGP tab
            'lgp_type': '',
            'lgp_period': 1,
            'lgp_file': '',
            # WHC Tab
            'whc_type': '',
            'whc_mm': 0.0,
            'whc_file': '',
            # Mask Tab
            'mask_type': '',
            'mask_file': '',
            # Precip/PET data
            'precip_type': '',
            'pet_type': '',
            'use_simulated_data': False,
            'ppt_dataset': '',
            'pet_dataset': '',
            'initial_period': '',
            'final_period': '',
            'start_year': 0,
            'cross_year': False,
            'available_files_list_ppt': [],
            'missing_files_list_ppt': [],
            'available_files_list_pet': [],
            'missing_files_list_pet': [],
            'output_file_list_wrsi': [],
            'output_file_list_lgp_phen': [],
            'output_file_list_swi': [],
            'output_file_wrsi_eos': '',
            'output_file_wrsi_anom_eos': '',
            'output_file_lgp_phen_eos': '',
            'output_file_swi_eos': '',
            'output_file_water_req_tot_eos': '',
            'output_file_max_water_def_eos': '',
            'output_file_water_def_tot_eos': '',
            'output_file_max_water_surp_eos': '',
            'output_file_water_surp_tot_eos': '',
            'output_file_aet_tot_eos': '',
            'output_file_sos': '',
            'output_file_sos_anom': '',
            'output_file_list_crop_stages': [],
            'type_of_sos_calculation': "3-Period Rain Threshold",
            'period_1_rainfall': 25,
            'period_2_3_rainfall': 20,
            'ignore_sos_clim': False,
            'max_periods_early': 2,
            'max_periods_late': 9,
            'recalculate_sos_after_crop_failure': False,
            'crop_failure_definition_for_restart': None,
            'crop_failure_definition_for_restart_percent': 30,
            'dont_restart_after': 40,
            'max_growing_periods_after_eos': 6,
            'exclude_areas_less_than': False,
            'exclude_areas_less_than_percent': 90
        }
        self.output_settings_path =\
            os.path.join(self.wrksp_setup.get_workspace(),
                         config.PROGRAM_SETTINGS,
                         config.GEOWRSI_OUT_SETTINGS_FILE)
        self.output_dic = geo_util.get_default_output_dic()
        self.output_dic['output_directory'] =\
            self.wrksp_setup.get_output_path()
        # used in the datachecks ---------------------------------------
        self.open_file_info = None
        # set up worker
        self.worker = None
        self.thread = None
        self.test_wrsi_color_file = \
            os.path.join(self.wrksp_setup.get_colors_path(),
                         config.WRSI_COLOR_FILE)
        self.test_swi_color_file = \
            os.path.join(self.wrksp_setup.get_colors_path(),
                         config.SWI_COLOR_FILE)
        self.test_lgp_phen_color_file = \
            os.path.join(self.wrksp_setup.get_colors_path(),
                         config.LGP_PHEN_COLOR_FILE)
        self.test_totals_color_file = \
            os.path.join(self.wrksp_setup.get_colors_path(),
                         config.TOTALS_COLOR_FILE)
        self.max_water_testing_color_file = \
            os.path.join(self.wrksp_setup.get_colors_path(),
                         config.MAX_COLOR_FILE)
        self.wrsi_anom_color_file = \
            os.path.join(self.wrksp_setup.get_colors_path(),
                         config.WRSI_ANOM_COLOR_FILE)
        self.sos_anom_dekads_color_file = \
            os.path.join(self.wrksp_setup.get_colors_path(),
                         config.SOS_ANOM_DEKAD_COLOR_FILE)
        self.sos_anom_pentads_color_file = \
            os.path.join(self.wrksp_setup.get_colors_path(),
                         config.SOS_ANOM_PENTAD_COLOR_FILE)

        # Setup Ui
        self.ui.progressBar.setValue(0)
        self.ui.cancelButton.setEnabled(False)
        self.closed = False  # A flag to indicate if close_event is triggered
        self.close_event = None
        load_error = self.initial_form_load()
        if load_error:
            self.ui.okButton.setEnabled(False)
        self.ui.okButton.clicked.connect(self.start_geowrsi)

    def initial_form_load(self):
        '''
        This function is called inside the init function in order to
        setup the default datasets for PPT and PET data, fill the years
        comboBox on GUI and read in the information from the
        config.GEOWRSI_SETTINGS_FILE file
        returns(bool) - if there was an error in loading data for form
        '''
        # read in the settings dictionary from the
        # config.GEOWRSI_SETTINGS_FILE file in ProgramSettings
        if self.fill_settings_dict():
            return True
        # Check for existing config.GEOWRSI_OUT_SETTINGS_FILE file
        # Fill it if it does, otherwise leave it as default settings
        self.output_dic = geo_util.set_output_dic_from_file(
            self.output_settings_path, self.output_dic)
        # Call load datasets function
        self.load_datasets()
        self.ui.lineEditPPT.setText(self.ds_dic_ppt['DATASETNAME'])
        self.ui.lineEditPPT.setEnabled(False)
        self.ui.lineEditPET.setText(self.ds_dic_pet['DATASETNAME'])
        self.ui.lineEditPET.setEnabled(False)
        self.ui.lineEditRegion.setText(self.set_dic['analysis_region'])
        self.ui.lineEditRegion.setEnabled(False)
        self.ui.lineEditCrop.setText(self.set_dic['crop_type'])
        self.ui.lineEditCrop.setEnabled(False)
        # Check each dataset was filled
        if not self.ds_dic_ppt:
            QMessageBox.warning(
                self,
                'Dataset missing!!',
                'No PPT dataset selected, '
                'or selected dataset does not exist!!',
                QMessageBox.Ok)
            return True
        if not self.ds_dic_pet:
            QMessageBox.warning(
                self,
                'Dataset missing!!',
                'No PET dataset selected, '
                'or selected dataset does not exist!!',
                QMessageBox.Ok)
            return True
        # Check datasets have same PERIODICITY
        if self.ds_dic_ppt['PERIODICITY'] != self.ds_dic_pet['PERIODICITY']:
            QMessageBox.information(self,
                                    u'Dataset Periodicity mismatch!!',
                                    'PPT and PET datasets do '
                                    'not use same periodicity',
                                    QMessageBox.Ok)
            self.ui.lineEdit.setText('Error - Dataset Mismatch')
            return True
        if self.ds_dic_ppt['DATASUFFIX'] != self.ds_dic_pet['DATASUFFIX']:
            QMessageBox.information(self,
                                    u'Dataset file type mismatch!!',
                                    'PPT and PET datasets do not '
                                    'use same file types',
                                    QMessageBox.Ok)
            self.ui.lineEdit.setText('Error - Dataset file type Mismatch')
            return True
        interval_dic = util.get_interval_dic(
            self.ds_dic_ppt['PERIODICITY'])
        self.ui.lineEdit.setText(self.ds_dic_ppt['PERIODICITY'])
        self.ui.lineEdit.setEnabled(False)
        # Based on PPT data to get list of all files in order
        # to get possible years
        input_file_list_ppt = \
            util.get_input_file_list(self.ds_dic_ppt)
        input_file_list_pet = \
            util.get_input_file_list(self.ds_dic_pet)
        if not input_file_list_ppt:
            self.ui.yearComboBox.clear()
            QMessageBox.warning(
                self,
                'Data missing!!',
                'No data available for PPT dataset '
                + self.ds_dic_ppt['DATASETNAME'] + '!!',
                QMessageBox.Ok)
            return True
        if not input_file_list_pet:
            self.ui.yearComboBox.clear()
            QMessageBox.warning(
                self,
                'Data missing!!',
                'No data available for PET dataset '
                + self.ds_dic_pet['DATASETNAME'] + '!!',
                QMessageBox.Ok)
            return True
        # Fill the year's combo box
        available_year_list_ppt = \
            util.extract_data_years(
                self.ds_dic_ppt, input_file_list_ppt)
        available_year_list_pet = \
            util.extract_data_years(
                self.ds_dic_pet, input_file_list_pet)
        available_year_list_both = list(
            set(available_year_list_ppt).intersection(
                set(available_year_list_pet)))
        year_end_period = len(interval_dic)
        mid_point = util.get_midpoint(self.ds_dic_ppt)
        util.fill_year_widget(self.ui.yearComboBox, False,
                              available_year_list_both,
                              1, year_end_period, mid_point)
        return False

    def load_datasets(self):
        '''
        This function will take the default PPT and default PET
        datasets, query them and set the ds_dic for each of them
        '''
        # load the dataset for PPT
        ds_info_ppt = DatasetsModel()
        if not ds_info_ppt.query_named_dataset(self.set_dic['ppt_dataset']):
            self.ds_dic_ppt = ds_info_ppt.get_ds_dictionary()
        else:
            self.ds_dic_ppt = None
        # load the dataset for PET
        ds_info_pet = DatasetsModel()
        if not ds_info_pet.query_named_dataset(self.set_dic['pet_dataset']):
            self.ds_dic_pet = ds_info_pet.get_ds_dictionary()
        else:
            self.ds_dic_pet = None

    def fill_settings_dict(self):
        '''
        This function will open the config.GEOWRSI_SETTINGS_FILE file in order
        to fill the set_dic with needed data from the
        geowrsi_settings_controller
        returns(bool) - err - if there was an error or not loading set_dic
        '''
        if os.path.exists(self.input_file_path):
            lines = []
            with open(self.input_file_path, "r") as f_obj:
                lines = f_obj.readlines()
            for line in lines:
                row = line.split(' ', 1)
                if row[1].strip() == "True":
                    self.set_dic[row[0]] = True
                elif row[1].strip() == "False":
                    self.set_dic[row[0]] = False
                else:
                    self.set_dic[row[0]] = row[1].strip()
            return False
        QMessageBox.information(
            self, config.GEOWRSI_SETTINGS_FILE + ' File: Path Not Found',
            'The path to the settings file cannot be found!', QMessageBox.Ok)
        return True

    def set_and_check_region_from_txt_file(self):
        '''
        This function will be called along with the datachecks at the start
        of the start_geowrsi function, it queries the named region that is
        read in from the settings dic and splits it from the comments about
        the region in order to get only the region name
        '''
        if not self.region_info.query_named_region(
                self.set_dic['analysis_region']):
            self.reg_dic = self.region_info.get_region_dictionary()
        else:
            QMessageBox.information(
                self, 'Region Not Found',
                'Error: The region from the ' + config.GEOWRSI_SETTINGS_FILE
                + ' file cannot be found in database!', QMessageBox.Ok)
            return True
        if self.ds_dic_ppt['PERIODICITY'] != self.reg_dic['PeriodType']:
            QMessageBox.information(
                self, 'Region vs Dataset Periodicity Mismatch',
                'Error: The region periodicity is '
                + self.reg_dic['PeriodType']
                + ', and datasets periodicity is '
                + self.ds_dic_ppt['PERIODICITY']
                + '.', QMessageBox.Ok)
            return True
        if self.check_files():
            return True
        self.set_dic['initial_period'] = self.reg_dic['InitialPeriod']
        self.set_dic['final_period'] = self.reg_dic['FinalPeriod']
        return False

    def check_files(self):
        '''
        This function is used to make sure that the selected and default file
        paths actually exists.
        '''
        flag = False
        if self.set_dic['sos_type'] == 'Selected':
            # Check that file path is not empty and that it exists
            if self.set_dic['sos_file'] == '':
                # Error - SOS type is Selected but no file has been chosen
                QMessageBox.information(
                    self, 'Missing selected sos file',
                    'Selected sos type: No file selected!', QMessageBox.Ok)
                flag = True
            elif not os.path.exists(self.set_dic['sos_file']):
                # Error - Path to selected file does not exist
                QMessageBox.information(
                    self, 'Selected SOS File: Path Not Found',
                    'The selected path to the SOS file cannot be found!',
                    QMessageBox.Ok)
                flag = True
            else:
                sos_params = qgs_util.extract_raster_file_params(
                    self.set_dic['sos_file'])
                sos_ok = \
                    qgs_util.check_dataset_vs_region_extents(
                        self, sos_params[0],
                        qgs_util.get_region_extent(self.reg_dic))
                if not sos_ok:
                    QMessageBox.information(
                        self, 'Selected SOS File: Error',
                        'This file has a region mismatch error',
                        QMessageBox.Ok)
                    flag = True
        elif self.set_dic['sos_type'] == 'Calculated':
            # Check that file path is not empty and that it exists
            if self.reg_dic['SOS'] == '':
                # Error - SOS type is Selected but no file has been chosen
                QMessageBox.information(
                    self, 'Missing caclulated sos file',
                    'Calculated sos type: Region SOS file not found!',
                    QMessageBox.Ok)
                flag = True
            elif not os.path.exists(self.reg_dic['SOS']):
                # Error - Path to selected file does not exist
                QMessageBox.information(
                    self, 'Calculated SOS File: Path Not Found',
                    'The path to the Region SOS file cannot be found!',
                    QMessageBox.Ok)
                flag = True
            else:
                sos_params = qgs_util.extract_raster_file_params(
                    self.reg_dic['SOS'])
                sos_ok = \
                    qgs_util.check_dataset_vs_region_extents(
                        self, sos_params[0],
                        qgs_util.get_region_extent(self.reg_dic))
                if not sos_ok:
                    QMessageBox.information(
                        self, 'Selected SOS File: Error',
                        'This file has a region mismatch error',
                        QMessageBox.Ok)
                    flag = True
        if self.set_dic['lgp_type'] == 'Selected' or\
                self.set_dic['lgp_type'] == 'Default':
            # Check that file path is not empty and that it exists
            if self.set_dic['lgp_file'] == '':
                # Error - SOS type is Selected but no file has been chosen
                QMessageBox.information(
                    self, 'Missing LGP File',
                    'LGP: No file selected!', QMessageBox.Ok)
                flag = True
            elif not os.path.exists(self.set_dic['lgp_file']):
                # Error - Path to selected file does not exist
                QMessageBox.information(
                    self, 'LGP File: Path Not Found',
                    'The path to the LGP file cannot be found!',
                    QMessageBox.Ok)
                flag = True
            else:  # Check file extents
                lgp_params = qgs_util.extract_raster_file_params(
                    self.set_dic['lgp_file'])
                lgp_ok = \
                    qgs_util.check_dataset_vs_region_extents(
                        self, lgp_params[0],
                        qgs_util.get_region_extent(self.reg_dic))
                if not lgp_ok:
                    QMessageBox.information(
                        self, 'LGP File: Error',
                        'This file has a region mismatch error',
                        QMessageBox.Ok)
                    flag = True
        if self.set_dic['whc_type'] == 'Selected' or\
                self.set_dic['whc_type'] == 'Default':
            # Check that file path is not empty and that it exists
            if self.set_dic['whc_file'] == '':
                # Error - SOS type is Selected but no file has been chosen
                QMessageBox.information(
                    self, 'Missing WHC File',
                    'WHC: No file selected!', QMessageBox.Ok)
                flag = True
            elif not os.path.exists(self.set_dic['whc_file']):
                # Error - Path to selected file does not exist
                QMessageBox.information(
                    self, 'WHC File: Path Not Found',
                    'The path to the WHC file cannot be found!',
                    QMessageBox.Ok)
                flag = True
            else:
                whc_params = qgs_util.extract_raster_file_params(
                    self.set_dic['whc_file'])
                whc_ok = \
                    qgs_util.check_dataset_vs_region_extents(
                        self, whc_params[0],
                        qgs_util.get_region_extent(self.reg_dic))
                if not whc_ok:
                    QMessageBox.information(
                        self, 'WHC File: Error',
                        'This file has a region mismatch error',
                        QMessageBox.Ok)
                    flag = True
        if self.set_dic['mask_type'] == 'Selected' or\
                self.set_dic['mask_type'] == 'Default':
            # Check that file path is not empty and that it exists
            if self.set_dic['mask_file'] == '':
                # Error - SOS type is Selected but no file has been chosen
                QMessageBox.information(
                    self, 'Missing Mask File',
                    'Mask: No file selected!', QMessageBox.Ok)
                flag = True
            elif not os.path.exists(self.set_dic['mask_file']):
                # Error - Path to selected file does not exist
                QMessageBox.information(
                    self, 'Mask File: Path Not Found',
                    'The path to the Mask file cannot be found!',
                    QMessageBox.Ok)
                flag = True
            else:
                mask_params = qgs_util.extract_raster_file_params(
                    self.set_dic['mask_file'])
                mask_ok = \
                    qgs_util.check_dataset_vs_region_extents(
                        self, mask_params[0],
                        qgs_util.get_region_extent(self.reg_dic))
                if not mask_ok:
                    QMessageBox.information(
                        self, 'Mask File: Error',
                        'This file has a region mismatch error',
                        QMessageBox.Ok)
                    flag = True
        return flag

    def get_ppt_pet_file_lists_wrsi(self, initial_period, final_period,
                                    start_year, cross_year):
        '''
        Takes the initialization data for region, year and datasets. Gets the
        file lists for the data encompassing the beginning of the period
        described by the region file, along with the 6 previous dekads of
        information before the period starts, and fills them into the
        available/missing file
        lists for future use
        params(int) - initial_period - initial period based on sos
        params(int) - final_period - final period based on sos+lgp
        params(int) - start_year - start year from the Run GUI
        params(bool) - cross_year - if it is a cross year then this is true
        '''
        all_periods_list_ppt = util.get_consecutive_periods_list(
            self.ds_dic_ppt)
        all_periods_list_pet = util.get_consecutive_periods_list(
            self.ds_dic_pet)
        idx_start = initial_period - 1
        idx_end = final_period - 1
        # Periodicity of both datasets is the same and was checked in form load
        intvl_dic = util.get_interval_dic(self.ds_dic_ppt['PERIODICITY'])
        end_yr_per = len(intvl_dic)
        year_lower = start_year
        if cross_year:
            year_higher = start_year + 1
        else:
            year_higher = start_year
        # Handles Pentad vs Dekad depending on what the dataset is setup as
        # Will reset the start period to 6 dekads or 12 pentads before to then
        # pass in as the periods to check for files
        if end_yr_per == 36:  # Handle as dekad
            # Index is 0-35 so if 6 then make it 0
            if idx_start >= 6:
                new_start = idx_start - 6
            else:
                # 5 needs to add 36 and sub 6, so add 30 set year back 1
                new_start = idx_start + 30
                year_lower -= 1
        elif end_yr_per == 72:  # Handle as pentad
            # index is 0-71, so if 12 it needs to be set to index of 0
            if idx_start >= 12:
                new_start = idx_start - 12
            else:
                new_start = idx_start + 60
                year_lower -= 1
        # Add on another year of data
        year_higher += 1
        # Reset the start_per set to the new index value
        # Set values to pass into the check_consecutive_process_inputs function
        # year_lower has needed start year
        # year_higher has needed end year
        start_period_ppt = all_periods_list_ppt[new_start]
        end_period_ppt = all_periods_list_ppt[idx_end]
        start_period_pet = all_periods_list_pet[new_start]
        end_period_pet = all_periods_list_pet[idx_end]
        # Call the util.check_consecutive_process_inputs function for PPT
        self.set_dic['available_files_list_ppt'] = []
        self.set_dic['missing_files_list_ppt'] = []
        self.set_dic['available_files_list_pet'] = []
        self.set_dic['missing_files_list_pet'] = []
        (_,
         self.set_dic['available_files_list_ppt'],
         self.set_dic['missing_files_list_ppt']) = \
            util.check_consecutive_process_inputs(
            self.ds_dic_ppt,
            year_lower, year_higher,
            start_period_ppt, end_period_ppt)
        # Call the util.check_consecutive_process_inputs function for PET
        (_,
         self.set_dic['available_files_list_pet'],
         self.set_dic['missing_files_list_pet']) = \
            util.check_consecutive_process_inputs(
            self.ds_dic_pet,
            year_lower, year_higher,
            start_period_pet, end_period_pet)

    def get_start_year(self):
        '''
        This function will get the year value from the comboBox
        '''
        self.set_dic['start_year'] = int(self.ui.yearComboBox.currentText())

    def get_crop_data(self):
        '''
        This function is used to set the crop_dic with needed crop values
        that will be used in the GeoWRSI tool later
        '''
        err = False
        crop_info = CropsModel()
        if not crop_info.query_named_crop(self.set_dic['crop_type']):
            self.crop_dic = crop_info.get_crop_dictionary()
        else:
            QMessageBox.information(
                self, 'Crop Not Found',
                'Error: The crop_type: "' + self.set_dic['crop_type'] +
                '" from the ' + config.GEOWRSI_SETTINGS_FILE +
                ' file cannot be found in database!', QMessageBox.Ok)
            err = True
        return err

    def start_geowrsi(self):
        '''
        This function will be called when the ok button is clicked for the
        GeoWrsi run GUI. It will take the year from the comboBox and then
        perform the calculations from there.
        '''
        # Initial setup to make sure all data is setup and correct ---------
        self.get_start_year()
        # Call set and check region function
        if self.set_and_check_region_from_txt_file():
            return
        self.set_dic['cross_year'] = \
            self.set_dic['initial_period'] > self.set_dic['final_period']
        # Call get_file_lists_wrsi function
        self.get_ppt_pet_file_lists_wrsi(self.set_dic['initial_period'],
                                         self.set_dic['final_period'],
                                         self.set_dic['start_year'],
                                         self.set_dic['cross_year'])
        if self.set_dic['missing_files_list_ppt'] or\
                self.set_dic['missing_files_list_pet']:
            self.ui.progressBar.setValue(0)
            QMessageBox.information(self,
                                    u'Missing files!!',
                                    str(self.set_dic['missing_files_list_ppt'])
                                    + str(self.set_dic[
                                        'missing_files_list_pet']),
                                    QMessageBox.Ok)
            QgsMessageLog.logMessage(u'Missing files!! \n PPT: ' +
                                     str(self.set_dic[
                                         'missing_files_list_ppt'])
                                     + "\n PET: "
                                     + str(self.set_dic[
                                         'missing_files_list_pet']),
                                     level=Qgis.Info)
            return
        if self.get_crop_data():
            return
        # Generate output file list
        self.set_dic['output_file_list_wrsi'] = self.get_output_file_list(
            "do")
        self.set_dic['output_file_list_swi'] = self.get_output_file_list(
            "dw")
        self.set_dic['output_file_list_lgp_phen'] = self.get_output_file_list(
            "dl")
        self.set_dic['output_file_wrsi_eos'] = self.get_output_file_name(
            "do", "_e")
        self.set_dic['output_file_swi_eos'] = self.get_output_file_name(
            "dw", "_e")
        self.set_dic['output_file_lgp_phen_eos'] = self.get_output_file_name(
            "dl", "_e")
        self.set_dic['output_file_water_req_tot_eos'] =\
            self.get_output_file_name("tr", "_e")
        self.set_dic['output_file_max_water_def_eos'] =\
            self.get_output_file_name("md", "_e")
        self.set_dic['output_file_water_def_tot_eos'] =\
            self.get_output_file_name("td", "_e")
        self.set_dic['output_file_max_water_surp_eos'] =\
            self.get_output_file_name("mx", "_e")
        self.set_dic['output_file_water_surp_tot_eos'] =\
            self.get_output_file_name("tx", "_e")
        self.set_dic['output_file_aet_tot_eos'] = self.get_output_file_name(
            "te", "_e")
        self.set_dic['output_file_wrsi_anom_eos'] = self.get_output_file_name(
            "er", "_e")
        self.set_dic['output_file_sos'] = self.get_output_file_name(
            "dt", "")
        self.set_dic['output_file_sos_anom'] = self.get_output_file_name(
            "dd", "")
        self.set_dic['output_file_list_crop_stages'] =\
            self.get_crop_stages_file_list()

        # Check for open files from output list, ppt, pet, mask, sos, whc, lgp
        self.open_file_info = qgs_util.get_open_files_info()
        self.check_for_open_files()
        '''
        Pass everything into the worker right here to run calcs
        Run actual GeoWRSI here passing in the stacked cube a pixel at a time
        Do this by creating a worker here and passing in the data we need to
        process through the WRSI
        '''
        # put info into worker
        self.worker = \
            GeoWRSIWorker(
                self.wrksp_setup,
                self.set_dic,
                self.reg_dic,
                self.crop_dic,
                self.output_dic)

        self.ui.closeButton.setEnabled(False)
        self.ui.cancelButton.setEnabled(True)
        self.ui.okButton.setEnabled(False)
        self.ui.yearComboBox.setEnabled(False)
        self.ui.cancelButton.clicked.connect(self.worker.kill)

        message_text = "WRSI calculating..."
        iface.mainWindow().statusBar().showMessage(message_text)
        # Start process in a new thread
        iface.messageBar().clearWidgets()
        self.thread = QtCore.QThread(self)
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.thread.start()
        self.worker.finished.connect(self.wrsi_finished)
        self.worker.error.connect(self.wrsi_error)
        self.worker.progress.connect(self.ui.progressBar.setValue)
        self.is_processing = True

    def wrsi_finished(self, return_value):
        '''
        Clean up the worker and thread
        '''
        self.is_processing = False
        self.thread.quit()
        self.thread.wait()
        logged_files = []
        if self.closed:
            self.thread_worker_cleanup()
            self.close_event.accept()
            QDialog.closeEvent(self, self.close_event)
            return
        if self.worker.killed:
            self.ui.progressBar.setValue(0)
            # msg = (u"WRSI aborted by user")
            iface.messageBar().pushCritical("WRSI", return_value[1])
            QgsMessageLog.logMessage(return_value[1], level=Qgis.Critical)
            iface.mainWindow().statusBar().showMessage(return_value[1], 3000)
        elif return_value is not None:
            msg = (u"Displaying output files")
            QgsMessageLog.logMessage(msg, level=Qgis.Info)
            iface.mainWindow().statusBar().showMessage(msg, 3000)

            lyr_tree = iface.layerTreeView()
            root = lyr_tree.currentGroupNode()
            if root is None:
                root = QgsProject.instance().layerTreeRoot()

            # Only ask about the animator if there is an output we animate
            if self.output_dic["save_every_wrsi"] or\
                self.output_dic["save_every_lgp_phenology"] or \
                    self.output_dic['save_every_soil_water_index']:
                ans = QMessageBox.question(
                    None, 'Animator',
                    ('Would you like the animator to be initialized?'),
                    QMessageBox.Yes, QMessageBox.No)
                animated = ans == QMessageBox.Yes
            else:
                animated = False

            if self.output_dic['eos_wrsi']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_wrsi_eos'],
                    self.test_wrsi_color_file)
                logged_files += [self.set_dic['output_file_wrsi_eos']]
            if self.output_dic['eos_wrsi_anomaly']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_wrsi_anom_eos'],
                    self.wrsi_anom_color_file)
                logged_files += [self.set_dic['output_file_wrsi_anom_eos']]
            if self.output_dic['eos_soil_water_index']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_swi_eos'],
                    self.test_swi_color_file)
                logged_files += [self.set_dic['output_file_swi_eos']]
            if self.output_dic['eos_lgp_phenology']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_lgp_phen_eos'],
                    self.test_lgp_phen_color_file)
                logged_files += [self.set_dic['output_file_lgp_phen_eos']]
            if self.output_dic['eos_water_req_totals']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_water_req_tot_eos'],
                    self.test_totals_color_file)
                logged_files += [self.set_dic['output_file_water_req_tot_eos']]
            if self.output_dic['eos_max_water_deficits']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_max_water_def_eos'],
                    self.max_water_testing_color_file)
                logged_files += [self.set_dic['output_file_max_water_def_eos']]
            if self.output_dic['eos_water_deficit_totals']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_water_def_tot_eos'],
                    self.test_totals_color_file)
                logged_files += [self.set_dic['output_file_water_def_tot_eos']]
            if self.output_dic['eos_max_water_surplus']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_max_water_surp_eos'],
                    self.max_water_testing_color_file)
                logged_files += [self.set_dic['output_file_max_water_surp_eos']]
            if self.output_dic['eos_water_surplus_totals']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_water_surp_tot_eos'],
                    self.test_totals_color_file)
                logged_files += [self.set_dic['output_file_water_surp_tot_eos']]
            if self.output_dic['eos_aet_totals']:
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_aet_tot_eos'],
                    self.test_totals_color_file)
                logged_files += [self.set_dic['output_file_aet_tot_eos']]
            qgs_util.display_raster_layer(
                self.set_dic['output_file_sos'],
                self.reg_dic['SOSColor'])
            logged_files += [self.set_dic['output_file_sos']]
            if self.reg_dic['PeriodType'] == 'Dekadal':
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_sos_anom'],
                    self.sos_anom_dekads_color_file)
                logged_files += [self.set_dic['output_file_sos_anom']]
            if self.reg_dic['PeriodType'] == 'Pentadal':
                qgs_util.display_raster_layer(
                    self.set_dic['output_file_sos_anom'],
                    self.sos_anom_pentads_color_file)
                logged_files += [self.set_dic['output_file_sos_anom']]

            if self.output_dic["save_every_wrsi"]:
                for file in self.set_dic['output_file_list_wrsi']:
                    qgs_util.display_raster_layer(
                        file, self.test_wrsi_color_file)
                logged_files += self.set_dic['output_file_list_wrsi']

            if self.output_dic["save_every_lgp_phenology"]:
                for file in self.set_dic['output_file_list_lgp_phen']:
                    qgs_util.display_raster_layer(
                        file, self.test_lgp_phen_color_file)
                logged_files += self.set_dic['output_file_list_lgp_phen']

            if self.output_dic["save_every_soil_water_index"]:
                for file in self.set_dic['output_file_list_swi']:
                    qgs_util.display_raster_layer(
                        file, self.test_swi_color_file)
                logged_files += self.set_dic['output_file_list_swi']
            if self.output_dic['save_crop_stage_totals_eos']:
                for file in self.set_dic['output_file_list_crop_stages']:
                    qgs_util.display_raster_layer(
                        file, self.test_totals_color_file)
                logged_files += self.set_dic['output_file_list_crop_stages']

            yr = self.set_dic['start_year']
            if self.set_dic['cross_year']:
                yr2 = yr + 1
                yr_str = str(yr) + "_" + str(yr2)
            else:
                yr_str = str(yr)
            out_suffix = " " + \
                self.output_dic['output_prefix'] + \
                self.set_dic['analysis_region'] + yr_str
            if self.output_dic["save_every_wrsi"]:
                if animated:
                    setup_animation(
                        self.set_dic['output_file_list_wrsi'],
                        self.ds_dic_ppt['PERIODICITY'])
                name = "WRSI" + out_suffix
                self.group_files(name, "output_file_list_wrsi", root)

            if self.output_dic["save_every_lgp_phenology"]:
                if animated:
                    setup_animation(
                        self.set_dic['output_file_list_lgp_phen'],
                        self.ds_dic_ppt['PERIODICITY'])
                name = "LGP Phenology" + out_suffix
                self.group_files(name, "output_file_list_lgp_phen", root)

            if self.output_dic["save_every_soil_water_index"]:
                if animated:
                    setup_animation(
                        self.set_dic['output_file_list_swi'],
                        self.ds_dic_ppt['PERIODICITY'])
                name = "Soil Water Index" + out_suffix
                self.group_files(name, "output_file_list_swi", root)

            if self.output_dic['save_crop_stage_totals_eos']:
                name = "Crop Stage Outputs" + out_suffix
                self.group_files(name, "output_file_list_crop_stages", root)

            if animated:
                animation_instructions = 'To use the animator set' +\
                    ' the step unit to "source timestamps" right after the ' +\
                    'animation range and then press the "set to full' +\
                    'range" directly to the left of the steps options. ' +\
                    'Now the user can navigate forward and backward through ' +\
                    'the periods by using the arrows on the temporal ' +\
                    'controller and switch which output is displayed by ' +\
                    'toggling on that group in the layer pannel.'
                QMessageBox.information(None, 'Animator instructions',
                                        animation_instructions, QMessageBox.Ok)

                QgsMessageLog.logMessage(animation_instructions,
                                         level=Qgis.Info)

                # Loop over the root things and hide every folder but one.
                for thing in QgsProject.instance().layerTreeRoot().children():
                    if self.output_dic["save_every_wrsi"]:
                        if thing.name() != "WRSI" + out_suffix:
                            thing.setItemVisibilityChecked(False)

                    elif self.output_dic["save_every_lgp_phenology"]:
                        if thing.name() != "LGP Phenology" + out_suffix:
                            thing.setItemVisibilityChecked(False)

                    elif self.output_dic["save_every_soil_water_index"]:
                        if thing.name() != "Soil Water Index" + out_suffix:
                            thing.setItemVisibilityChecked(False)

                    else:
                        thing.setItemVisibilityChecked(False)
                set_start_end(
                    self.set_dic["output_file_list_wrsi"],
                    self.ds_dic_ppt['PERIODICITY'])

            lyr_tree.collapseAllNodes()
            lyr_tree.setCurrentLayer(None)
            self.ui.progressBar.setValue(100)
            QgsMessageLog.logMessage(return_value[1], level=Qgis.Info)
            iface.mainWindow().statusBar().showMessage(return_value[1], 3000)
        else:
            # notify of an error
            msg = (u'Error: WRSI did not finish')
            iface.messageBar().pushCritical('WRSI', msg)
            QgsMessageLog.logMessage(msg, level=Qgis.Critical)
            iface.mainWindow().statusBar().showMessage(msg, 3000)

        self.ui.closeButton.setEnabled(True)
        self.ui.cancelButton.setEnabled(False)
        self.ui.okButton.setEnabled(True)
        self.ui.yearComboBox.setEnabled(True)
        self.thread_worker_cleanup()
        set_dic_log_string = []
        for pair in self.set_dic.items():
            if "output_file" not in pair[0]:
                set_dic_log_string += log_util.log_string(pair[0], [pair[1]])
        log_util.add_log("WRSI",
                         set_dic_log_string,
                         logged_files)

        all_done = util.notify_process_completed('WRSI')
        if all_done:
            self.ui.closeButton.click()

    def wrsi_error(self, err, ex_str):
        '''
        Function to log any errors from the object and thread.
        Arguments:
        err -- Not used
        ex_str -- Exception info
        '''
        self.is_processing = False
        QgsMessageLog.logMessage(ex_str, level=Qgis.Critical)
        QMessageBox.critical(self, 'Error!', ex_str, QMessageBox.Ok)
        iface.mainWindow().statusBar().showMessage(ex_str)

    def closeEvent(self, a_0: QtGui.QCloseEvent) -> None:
        '''
        Close event used to handle a case where user closes the form
        while the process is running.
        '''
        self.closed = True
        if self.worker:
            self.worker.kill()
        self.close_event = a_0
        if self.is_processing:
            self.close_event.ignore()
        else:
            self.close_event.accept()

    def thread_worker_cleanup(self):
        """
        Cleanup thread if it exists.
        """
        if self.thread is not None:
            self.thread.deleteLater()

    def get_output_file_list(self, str_name):
        '''
        Function to generate the list of output names for the wrsi
        outputs should only be from region initial period to the
        region final period + 6 or 12 depending on Dekad or Pentad
        params(string) - str_name - pass in "do", "md", or
                        "mx", etc.
        returns(list) - output_list - returns the output file list
                        that was generated
        '''
        output_list = []
        start_per = self.set_dic['initial_period']
        yr = self.set_dic['start_year']
        end_per = self.set_dic['final_period']
        if self.reg_dic['PeriodType'] == 'Dekadal':
            if start_per > end_per:
                end_per += 36
            # Add 6 to include 6 dekads after final_period
            end_per += 6
            last_per = 36
            period = "_d"

        elif self.reg_dic['PeriodType'] == 'Pentadal':
            if start_per > end_per:
                end_per += 72
            # Add 12 to include 12 pentads after final_period
            end_per += 12
            last_per = 72
            period = "_p"

        # Add all files to the output_directory path in the output options
        out_folder_path = os.path.join(self.output_dic['output_directory'])
        if not os.path.exists(out_folder_path):
            os.makedirs(out_folder_path)

        per = start_per
        for _ in range(start_per, end_per + 1):
            if per > last_per:
                per = 1
                yr += 1
            out_file = os.path.join(out_folder_path,
                                    self.output_dic['output_prefix']
                                    + "w" + str(yr) + str_name
                                    + period + str(per).zfill(2)
                                    + self.ds_dic_ppt['DATASUFFIX'])
            output_list.append(out_file)
            per += 1

        return output_list

    def get_crop_stages_file_list(self):
        '''
        Function to generate the file list for the crop stages file lists
        returns(list) crop_stages_list - list of all the crop stages files
        '''
        crop_stages_list = []
        stages = ["i", "v", "f", "r"]
        names = ["e", "r", "x", "d"]

        for i in names:
            for j in stages:
                out_file = self.get_output_file_name(j + i, "")
                crop_stages_list.append(out_file)

        return crop_stages_list

    def get_output_file_name(self, str_name, suffix):
        '''
        Function to generate the output names for the wrsi outputs
        either current or eos
        params(string) - str_name - pass in "do", "md", or
                        "mx", etc.
        params(string) - suffix - eos or current, _e or _c
        returns(list) - out_file - returns the output file name
                        that was generated
        '''
        yr = self.set_dic['start_year']
        if self.set_dic['cross_year']:
            yr2 = yr + 1
            yr_str = str(yr) + "_" + str(yr2)
        else:
            yr_str = str(yr)

        # Add all files to the output_directory path in the output options
        out_folder_path = os.path.join(self.output_dic['output_directory'])
        if not os.path.exists(out_folder_path):
            os.makedirs(out_folder_path)

        out_file = os.path.join(out_folder_path,
                                self.output_dic['output_prefix']
                                + "w" + yr_str + str_name
                                + suffix + self.ds_dic_ppt['DATASUFFIX'])
        return out_file

    def check_for_open_files(self):
        '''
        Function that will check to make sure none of the needed files
        are open in qgis
        '''
        for name in self.set_dic['output_file_list_wrsi']:
            b_name = os.path.basename(name)
            self.check_loaded_map(b_name)
        for name in self.set_dic['output_file_list_swi']:
            b_name = os.path.basename(name)
            self.check_loaded_map(b_name)
        for name in self.set_dic['output_file_list_lgp_phen']:
            b_name = os.path.basename(name)
            self.check_loaded_map(b_name)
        for name in self.set_dic['available_files_list_ppt']:
            b_name = os.path.basename(name)
            self.check_loaded_map(b_name)
        for name in self.set_dic['available_files_list_pet']:
            b_name = os.path.basename(name)
            self.check_loaded_map(b_name)
        b_name = os.path.basename(self.set_dic['output_file_wrsi_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(self.set_dic['output_file_wrsi_anom_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(self.set_dic['output_file_swi_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(self.set_dic['output_file_lgp_phen_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(
            self.set_dic['output_file_water_req_tot_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(
            self.set_dic['output_file_max_water_def_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(
            self.set_dic['output_file_water_def_tot_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(
            self.set_dic['output_file_max_water_surp_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(
            self.set_dic['output_file_water_surp_tot_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(self.set_dic['output_file_aet_tot_eos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(self.set_dic['output_file_sos'])
        self.check_loaded_map(b_name)
        b_name = os.path.basename(self.set_dic['output_file_sos_anom'])
        self.check_loaded_map(b_name)
        for name in self.set_dic['output_file_list_crop_stages']:
            b_name = os.path.basename(name)
            self.check_loaded_map(b_name)
        if os.path.exists(self.set_dic['sos_file']):
            b_name = os.path.basename(self.set_dic['sos_file'])
            self.check_loaded_map(b_name)
        if os.path.exists(self.set_dic['lgp_file']):
            b_name = os.path.basename(self.set_dic['lgp_file'])
            self.check_loaded_map(b_name)
        if os.path.exists(self.set_dic['whc_file']):
            b_name = os.path.basename(self.set_dic['whc_file'])
            self.check_loaded_map(b_name)
        if os.path.exists(self.set_dic['mask_file']):
            b_name = os.path.basename(self.set_dic['mask_file'])
            self.check_loaded_map(b_name)

    def check_loaded_map(self, basename):
        '''
        Check a basename to see if it's in the map panel and notify
        if it is.  Will remove the specified map layer.
        Arguments:
        basename -- Basename as a string
        '''
        root_ext = os.path.splitext(basename)
        for entry in self.open_file_info:
            if root_ext[0] in entry[1]:
                qgs_util.notify_loaded_file(self, entry[1], entry[0])

    def group_files(self, group_name, file_list, root):
        '''
        Set up a layer group to output the passed list into
        params(string) - group_name - the name of the group
        params(list) - file_list - list of layernames to go into the group
        params(layerTreeRoot) - root - The layer tree root used for grouping
        '''
        if root.findGroup(group_name) is None:
            root.addGroup(group_name)
        parent = root.findGroup(group_name)

        for layer in root.findLayers():
            if QgsLayerTree.isLayer(layer):
                lname = layer.name()
                for file in self.set_dic[file_list]:
                    if lname == os.path.splitext(os.path.basename(file))[0]:
                        clone = layer.clone()
                        parent.insertChildNode(0, clone)
                        root.removeChildNode(layer)
