'''
/***************************************************************************
Name       :  geowrsi_worker.py
Description:  Worker for GeoWRSI Tool
copyright  :  (C) 2022-2023 by FEWS
email      :  dhackman@contractor.usgs.gov
Author     :  Derek Hackman
Modified   :  mm/dd/yyyy - descripton - Description

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
'''
import os

import numpy as np

from PyQt5 import QtCore

from qgis.core import Qgis, QgsMessageLog

from fews_tools.utilities import geoclim_gdal_utilities as gdal_util
from fews_tools.utilities import geoclim_utilities as util
from fews_tools.utilities import geowrsi_utilities as geo_util

DEKAD_SHIFT = 6
PENTAD_SHIFT = 12


class GeoWRSIWorker(QtCore.QObject):
    '''
    GeoWRSI worker class
    '''

    def __init__(self, wrksp_setup, set_dic, reg_dic, crop_dic, output_dic):
        QtCore.QObject.__init__(self)
        self.wrksp_setup = wrksp_setup
        self.set_dic = set_dic
        self.reg_dic = reg_dic
        self.crop_dic = crop_dic
        self.output_dic = output_dic
        self.output_file_path =\
            self.wrksp_setup.get_output_path()
        self.killed = False

    def kill(self):
        """
        Set kill flog to true to stop processing.
        """
        self.killed = True

    def run(self):
        '''
        Run the geowrsi function
        '''
        ret_tuple = None
        err = False
        try:
            # Clear temp folder to makes sure all files
            # are resampled to region
            QgsMessageLog.logMessage(u"Clearing the temp folder",
                                     level=Qgis.Info)
            util.remove_temp_files(self.output_file_path)

            # Start resampling the needed files
            QgsMessageLog.logMessage(u"Resampling and gathering data",
                                     level=Qgis.Info)
            # Set the resampled mask
            resampled_mask, _ = \
                gdal_util.resample_input_files(self.output_file_path,
                                               self.set_dic['mask_file'],
                                               [], self.reg_dic)
            if self.killed is True:
                raise KeyboardInterrupt
            if not os.path.exists(resampled_mask):
                QgsMessageLog.logMessage(u'Mask File Error!! '
                                         'Error resampling the mask file',
                                         level=Qgis.Info)
                raise RuntimeError
            # Set resampled Clim WRSI
            clim_wrsi_cube =\
                geo_util.fill_data_cube_from_files(self.reg_dic,
                                                   [self.reg_dic["WRSI"]],
                                                   self.output_file_path)
            # Set resampled Clim SOS
            clim_sos_cube =\
                geo_util.fill_data_cube_from_files(self.reg_dic,
                                                   [self.reg_dic["SOS"]],
                                                   self.output_file_path)
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(5)
            # Set the PPT data cube and fill it with data
            ppt_data_cube =\
                geo_util.fill_data_cube_from_files(self.reg_dic,
                                                   self.set_dic['available_files_list_ppt'],
                                                   self.output_file_path)
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(20)
            # Set the PET data cube and fill it with data
            pet_data_cube =\
                geo_util.fill_data_cube_from_files(self.reg_dic,
                                                   self.set_dic['available_files_list_pet'],
                                                   self.output_file_path)
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(40)
            # Create lgp data Cube
            lgp_data_cube = geo_util.fill_lgp(self.set_dic, self.reg_dic,
                                              resampled_mask,
                                              self.output_file_path)
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(45)
            # Create sos data cube
            sos_data_cube = geo_util.fill_sos(self.set_dic, self.reg_dic,
                                              resampled_mask,
                                              self.output_file_path)
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(50)
            # Create WHC data Cube
            whc_data_cube = geo_util.fill_whc(self.set_dic, self.reg_dic,
                                              resampled_mask,
                                              self.output_file_path)
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(55)
            # Create Mask data cube
            mask_data_cube = geo_util.fill_mask(resampled_mask)
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(60)
            # Stack the data cubes
            cubes_arr = []
            cubes_arr.append(ppt_data_cube)
            cubes_arr.append(pet_data_cube)
            cubes_arr.append(lgp_data_cube)
            cubes_arr.append(whc_data_cube)
            cubes_arr.append(sos_data_cube)
            cubes_arr.append(mask_data_cube)
            stacked_cube = geo_util.stack_cubes(cubes_arr)
            # This extra_periods variable is used for calc_sos if needed and
            # for writing out to files later on so define it every time
            if self.reg_dic['PeriodType'] == 'Dekadal':
                extra_periods = DEKAD_SHIFT
            elif self.reg_dic['PeriodType'] == 'Pentadal':
                extra_periods = PENTAD_SHIFT
            else:
                QgsMessageLog.logMessage(u'Incorrect Region Period Type!! '
                                         'Must be "Dekadal" or "Pentadal"',
                                         level=Qgis.Info)
                raise RuntimeError
            if self.set_dic['sos_type'] == 'Calculated':
                QgsMessageLog.logMessage(u"Calculating SOS",
                                         level=Qgis.Info)
                stacked_cube = geo_util.calc_sos(
                    stacked_cube, self.reg_dic, self.set_dic, extra_periods)
            periods, _, _ = ppt_data_cube.shape
            QgsMessageLog.logMessage(u"Begin calculating WRSI",
                                     level=Qgis.Info)
            # Call geo_util.run_geowrsi function to get wrsi
            (wrsi_cube, swi_cube, lgp_phen_cube,
             eos_output_cube, crop_stages_cube) =\
                np.apply_along_axis(geo_util.run_geowrsi, 0,
                                    stacked_cube,
                                    self.crop_dic, periods,
                                    self.set_dic['initial_period'],
                                    self.set_dic['final_period'],
                                    self.set_dic['cross_year'],
                                    self.reg_dic['PeriodType'])
            if self.killed is True:
                raise KeyboardInterrupt

            # Calculate WRSI Anomaly
            mask = gdal_util.get_inverted_mask_array(resampled_mask)
            wrsi_arr = eos_output_cube[0]
            wrsi_anom = geo_util.anomaly_calc(wrsi_arr,
                                              self.reg_dic,
                                              mask,
                                              clim_wrsi_cube[0],
                                              "WRSI",
                                              self.set_dic['cross_year'])
            wrsi_anom[wrsi_anom == -9999] = 255

            # Calculate SOS anomaly
            sos_arr = stacked_cube[-2]
            sos_anom = geo_util.anomaly_calc(sos_arr,
                                             self.reg_dic,
                                             mask,
                                             clim_sos_cube[0],
                                             "SOS",
                                             self.set_dic['cross_year'])

            self.progress.emit(75)
            QgsMessageLog.logMessage(u"Writing output files",
                                     level=Qgis.Info)

            # Setup the parameters for writing to files
            _, row_ct, col_ct, geoxform, _ = \
                gdal_util.get_geotiff_info(resampled_mask)
            gd_data_type = gdal_util.TYPE_DIC["Float32"]["GDAL"]

            # Write the wrsi_cube out to a file for each needed output period
            # this will allow first 6 dekads or 12 pentads from data cube that
            # represent the preseason to be skipped and the extra periods
            # after the eos + 6 to be ignored
            if self.output_dic['save_every_wrsi']:
                for index, file in enumerate(self.set_dic[
                        'output_file_list_wrsi'],
                        extra_periods):
                    err = gdal_util.write_file(file, wrsi_cube[index], col_ct,
                                               row_ct, geoxform, gd_data_type)
                    if err is True:
                        raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(80)

            # Write the swi_cube out to a file for each needed output period
            # this will allow first 6 dekads or 12 pentads from data cube that
            # represent the preseason to be skipped and the extra periods
            # after the eos + 6 to be ignored
            if self.output_dic['save_every_soil_water_index']:
                for index, file in enumerate(self.set_dic[
                        'output_file_list_swi'],
                        extra_periods):
                    err = gdal_util.write_file(file, swi_cube[index], col_ct,
                                               row_ct, geoxform, gd_data_type)
                    if err is True:
                        raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(83)

            # Write the lgp_phen_cube out to file for each needed output period
            # this will allow first 6 dekads or 12 pentads from data cube that
            # represent the preseason to be skipped and the extra periods
            # after the eos + 6 to be ignored
            if self.output_dic['save_every_lgp_phenology']:
                for index, file in enumerate(self.set_dic[
                        'output_file_list_lgp_phen'],
                        extra_periods):
                    err = gdal_util.write_file(file, lgp_phen_cube[index],
                                               col_ct, row_ct,
                                               geoxform, gd_data_type)
                    if err is True:
                        raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(86)

            # Write out the wrsi_eos 2d array to the eos file
            if self.output_dic['eos_wrsi']:
                err = gdal_util.write_file(self.set_dic[
                    'output_file_wrsi_eos'],
                    eos_output_cube[0], col_ct, row_ct,
                    geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            # Write out the wrsi_eos 2d array to the eos file
            if self.output_dic['eos_wrsi_anomaly']:
                err = gdal_util.write_file(self.set_dic[
                    'output_file_wrsi_anom_eos'],
                    wrsi_anom, col_ct, row_ct,
                    geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            err = gdal_util.write_file(self.set_dic['output_file_sos'],
                                       sos_arr, col_ct, row_ct,
                                       geoxform, gd_data_type)
            if err is True:
                raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            err = gdal_util.write_file(self.set_dic['output_file_sos_anom'],
                                       sos_anom, col_ct, row_ct,
                                       geoxform, gd_data_type)
            if err is True:
                raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            # Write out the lgp_phen_eos 2d array to the eos file
            if self.output_dic['eos_lgp_phenology']:
                err = gdal_util.write_file(self.set_dic[
                    'output_file_lgp_phen_eos'],
                    eos_output_cube[2], col_ct, row_ct,
                    geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            # Write out the swi_eos 2d array to the eos file
            if self.output_dic['eos_soil_water_index']:
                err = gdal_util.write_file(self.set_dic['output_file_swi_eos'],
                                           eos_output_cube[3], col_ct, row_ct,
                                           geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            # Write out the water_req_total 2d array to the eos file
            if self.output_dic['eos_water_req_totals']:
                err = gdal_util.write_file(self.set_dic[
                    'output_file_water_req_tot_eos'],
                    eos_output_cube[4], col_ct, row_ct,
                    geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            # Write out the max_water_deficit 2d array to the eos file
            if self.output_dic['eos_max_water_deficits']:
                err = gdal_util.write_file(self.set_dic[
                    'output_file_max_water_def_eos'],
                    eos_output_cube[5], col_ct, row_ct,
                    geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            # Write out the water_deficit_total 2d array to the eos file
            if self.output_dic['eos_water_deficit_totals']:
                err = gdal_util.write_file(self.set_dic[
                    'output_file_water_def_tot_eos'],
                    eos_output_cube[6], col_ct, row_ct,
                    geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            # Write out the max_water_surplus 2d array to the eos file
            if self.output_dic['eos_max_water_surplus']:
                err = gdal_util.write_file(self.set_dic[
                    'output_file_max_water_surp_eos'],
                    eos_output_cube[7], col_ct, row_ct,
                    geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            # Write out the water_surplus_total 2d array to the eos file
            if self.output_dic['eos_water_surplus_totals']:
                err = gdal_util.write_file(self.set_dic[
                    'output_file_water_surp_tot_eos'],
                    eos_output_cube[8], col_ct, row_ct,
                    geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt

            # Write out the aet_total 2d array to the eos file
            if self.output_dic['eos_aet_totals']:
                err = gdal_util.write_file(self.set_dic[
                    'output_file_aet_tot_eos'],
                    eos_output_cube[9], col_ct, row_ct,
                    geoxform, gd_data_type)
                if err is True:
                    raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt
            self.progress.emit(90)

            if self.output_dic['save_crop_stage_totals_eos']:
                for index, file in enumerate(self.set_dic[
                        'output_file_list_crop_stages']):
                    err = gdal_util.write_file(file, crop_stages_cube[index],
                                               col_ct, row_ct,
                                               geoxform, gd_data_type)
                    if err is True:
                        raise RuntimeError

            if err is True:
                raise RuntimeError
            if self.killed is True:
                raise KeyboardInterrupt
            if self.killed is False:
                self.progress.emit(95)
                ret_tuple = (0, u"WRSI completed")
            if err is True:
                raise RuntimeError
        # exit with appropriate message on killed (KeyboardInterrupt)
        except KeyboardInterrupt:
            self.progress.emit(0)
            ret_tuple = (0, u"WRSI aborted by user")
        # forward any execeptions upstream
        except BaseException as exc:
            self.error.emit(exc, u"Unspecified error in WRSI")
        self.finished.emit(ret_tuple)

    finished = QtCore.pyqtSignal(object)

    error = QtCore.pyqtSignal(Exception, str)

    progress = QtCore.pyqtSignal(float)
